;; Generated package description from clojure-ts-mode.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "clojure-ts-mode" "0.6.0.0.20260708.47" "Major mode for Clojure code" '((emacs "30.1")) :commit "2f2da7e60eb8e1775d902586d332d73ddba8e24a" :maintainer '("Bozhidar Batsov" . "bozhidar@batsov.dev") :keywords '("languages" "clojure" "clojurescript" "lisp") :url "http://github.com/clojure-emacs/clojure-ts-mode")
